/*
 * COPYRIGHT:        See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System Driver for WinNT/2K/XP
 * FILE:             Ext2.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:         http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ext2fs.h"

/* GLOBALS ***************************************************************/

extern PEXT2_GLOBAL Ext2Global;

/* DEFINITIONS *************************************************************/

#ifdef ALLOC_PRAGMA
#pragma alloc_text(PAGE, Ext2LoadSuper)
#pragma alloc_text(PAGE, Ext2SaveSuper)

#pragma alloc_text(PAGE, Ext2LoadGroup)
#pragma alloc_text(PAGE, Ext2SaveGroup)

#pragma alloc_text(PAGE, Ext2GetInodeLba)
#pragma alloc_text(PAGE, Ext2LoadInode)
#pragma alloc_text(PAGE, Ext2SaveInode)

#pragma alloc_text(PAGE, Ext2LoadBlock)
#pragma alloc_text(PAGE, Ext2SaveBlock)

#pragma alloc_text(PAGE, Ext2SaveBuffer)

#pragma alloc_text(PAGE, Ext2GetBlock)
#pragma alloc_text(PAGE, Ext2BlockMap)

#pragma alloc_text(PAGE, Ext2BuildBDL)

#pragma alloc_text(PAGE, Ext2NewBlock)
#pragma alloc_text(PAGE, Ext2FreeBlock)

#pragma alloc_text(PAGE, Ext2ExpandBlock)
#pragma alloc_text(PAGE, Ext2ExpandInode)

#pragma alloc_text(PAGE, Ext2NewInode)
#pragma alloc_text(PAGE, Ext2FreeInode)

#pragma alloc_text(PAGE, Ext2AddEntry)
#pragma alloc_text(PAGE, Ext2RemoveEntry)

#pragma alloc_text(PAGE, Ext2TruncateBlock)

#pragma alloc_text(PAGE, Ext2AddMcbEntry)
#pragma alloc_text(PAGE, Ext2RemoveMcbEntry)
#pragma alloc_text(PAGE, Ext2LookupMcbEntry)

#endif

/* FUNCTIONS ***************************************************************/

NTSTATUS
Ext2LoadSuper(IN PEXT2_VCB      Vcb,
              IN BOOLEAN        bVerify,
              OUT PEXT2_SUPER_BLOCK * Sb)
{
    NTSTATUS          Status;
    PEXT2_SUPER_BLOCK Ext2Sb = NULL;
    
    Ext2Sb = (PEXT2_SUPER_BLOCK)
                ExAllocatePoolWithTag(
                    PagedPool,
                    SUPER_BLOCK_SIZE,
                    EXT2_SB_MAGIC
                );
    if (!Ext2Sb) {
        Status = STATUS_INSUFFICIENT_RESOURCES;
        goto errorout;
    }

    Status = Ext2ReadDisk(
                Vcb,
                (ULONGLONG) SUPER_BLOCK_OFFSET,
                SUPER_BLOCK_SIZE,
                (PVOID) Ext2Sb,
                bVerify );

    if (!NT_SUCCESS(Status)) {

        Ext2Print((DBG_ERROR, "Ext2ReadDisk: Read Block Device error.\n"));

        ExFreePoolWithTag(Ext2Sb, EXT2_SB_MAGIC);
        Ext2Sb = NULL;
    }

errorout:

    *Sb = Ext2Sb;
    return Status;
}


BOOLEAN
Ext2SaveSuper(  IN PEXT2_IRP_CONTEXT    IrpContext,
                IN PEXT2_VCB            Vcb )
{
    LONGLONG    Offset;
    BOOLEAN     bRet;

    Offset = (LONGLONG) SUPER_BLOCK_OFFSET;

    bRet = Ext2SaveBuffer( IrpContext,
                           Vcb,
                           Offset,
                           SUPER_BLOCK_SIZE,
                           Vcb->SuperBlock );

    if (IsFlagOn(Vcb->Flags, VCB_FLOPPY_DISK)) {
        Ext2StartFloppyFlushDpc(Vcb, NULL, NULL);
    }

    return bRet;
}


BOOLEAN
Ext2LoadGroup(IN PEXT2_VCB Vcb)
{
    ULONG       Size;
    PVOID       Buffer;
    LONGLONG    Lba = 0;
    NTSTATUS    Status;

    PEXT2_SUPER_BLOCK sb;

    sb = Vcb->SuperBlock;

    Vcb->BlockSize  = EXT2_MIN_BLOCK << sb->s_log_block_size;
    Vcb->SectorBits = Ext2Log2(SECTOR_SIZE);
    ASSERT(BLOCK_BITS == Ext2Log2(BLOCK_SIZE));

    Vcb->NumOfGroups = (sb->s_blocks_count - sb->s_first_data_block +
        sb->s_blocks_per_group - 1) / sb->s_blocks_per_group;

    Size = sizeof(EXT2_GROUP_DESC) * Vcb->NumOfGroups;

    if (Vcb->BlockSize == EXT2_MIN_BLOCK) {
        Lba = (LONGLONG)2 * Vcb->BlockSize;
    }

    if (Vcb->BlockSize > EXT2_MIN_BLOCK) {
        Lba = (LONGLONG) (Vcb->BlockSize);
    }

    if (Lba == 0) {
        return FALSE;
    }

    Buffer = ExAllocatePoolWithTag(PagedPool, Size, EXT2_GD_MAGIC);
    if (!Buffer) {
        Ext2Print((DBG_ERROR, "Ext2LoadSuper: no enough memory.\n"));
        return FALSE;
    }

    Ext2Print((DBG_INFO, "Ext2LoadGroup: Lba=%I64xh Size=%xh\n", Lba, Size));

    Status = Ext2ReadDisk(  Vcb,
                            Lba,
                            Size,
                            Buffer,
                            FALSE );

    if (!NT_SUCCESS(Status)) {
        ExFreePoolWithTag(Buffer, EXT2_GD_MAGIC);
        Buffer = NULL;

        return FALSE;
    }

/*
    bPinned = CcPinRead(
                Vcb->StreamObj,
                Lba, 
                Size,
                PIN_WAIT,
                &(Vcb->GroupDescBcb),
                &(Buffer));

    if (!bPinned)
    {
        Vcb->GroupDesc = NULL;
        return FALSE;
    }
*/

    Vcb->GroupDesc = (PEXT2_GROUP_DESC) Buffer;

    return TRUE;
}


BOOLEAN
Ext2SaveGroup(  IN PEXT2_IRP_CONTEXT    IrpContext,
                IN PEXT2_VCB            Vcb,
                IN ULONG                Group )
{
    LONGLONG    Offset;
    BOOLEAN     bRet;

    if (Vcb->BlockSize == EXT2_MIN_BLOCK) {

        Offset = (LONGLONG) (2 * Vcb->BlockSize);

    } else {

        Offset = (LONGLONG) (Vcb->BlockSize);
    }

    Offset += ((LONGLONG) sizeof(struct ext2_group_desc) * Group);

    bRet = Ext2SaveBuffer(
                    IrpContext,
                    Vcb,
                    Offset,
                    sizeof(struct ext2_group_desc),
                    &(Vcb->GroupDesc[Group]) );

    if (IsFlagOn(Vcb->Flags, VCB_FLOPPY_DISK)) {
        Ext2StartFloppyFlushDpc(Vcb, NULL, NULL);
    }

    return bRet;
}          

BOOLEAN
Ext2GetInodeLba (IN PEXT2_VCB   Vcb,
         IN  ULONG  inode,
         OUT PLONGLONG offset)
{
    LONGLONG loc;

    if (inode < 1 || inode > INODES_COUNT)  {
        Ext2Print((DBG_ERROR, "Ext2GetInodeLba: Inode value %xh is invalid.\n",inode));
        *offset = 0;
        return FALSE;
    }

    loc = (Vcb->GroupDesc[(inode - 1) / INODES_PER_GROUP ].bg_inode_table);
    loc = loc * Vcb->BlockSize;
    loc = loc + ((inode - 1) % INODES_PER_GROUP) * sizeof(EXT2_INODE);

    *offset = loc;

    return TRUE;
}

BOOLEAN
Ext2LoadInode (IN PEXT2_VCB Vcb,
           IN ULONG       inode,
           IN PEXT2_INODE Inode)
{
    IO_STATUS_BLOCK     IoStatus;
    LONGLONG            Offset; 

    if (!Ext2GetInodeLba(Vcb, inode, &Offset))  {
        Ext2Print((DBG_ERROR, "Ext2LoadInode: error get inode(%xh)'s addr.\n", inode));
        return FALSE;
    }

    if (!Ext2CopyRead(
            Vcb->StreamObj,
            (PLARGE_INTEGER)&Offset,
            sizeof(EXT2_INODE),
            PIN_WAIT,
            (PVOID)Inode,
            &IoStatus )) {
        return FALSE;
    }

    if (!NT_SUCCESS(IoStatus.Status)) {
        return FALSE;
    }

    return TRUE;
}


BOOLEAN
Ext2SaveInode ( IN PEXT2_IRP_CONTEXT IrpContext,
                IN PEXT2_VCB Vcb,
                IN ULONG Inode,
                IN PEXT2_INODE Ext2Inode)
{
    LONGLONG        Offset = 0;
    BOOLEAN         bRet;

    Ext2Print((DBG_INFO, "Ext2SaveInode: Saving Inode %xh: Mode=%xh Size=%xh\n",
                         Inode, Ext2Inode->i_mode, Ext2Inode->i_size));

    if (!Ext2GetInodeLba(Vcb, Inode, &Offset))  {
        Ext2Print((DBG_ERROR, "Ext2SaveInode: error get inode(%xh)'s addr.\n", Inode));
        return FALSE;
    }

    bRet = Ext2SaveBuffer(IrpContext, Vcb, Offset, sizeof(EXT2_INODE), Ext2Inode);

    if (IsFlagOn(Vcb->Flags, VCB_FLOPPY_DISK)) {
        Ext2StartFloppyFlushDpc(Vcb, NULL, NULL);
    }

    return bRet;
}


BOOLEAN
Ext2LoadBlock (IN PEXT2_VCB Vcb,
           IN ULONG     dwBlk,
           IN PVOID     Buffer )
{
    IO_STATUS_BLOCK     IoStatus;
    LONGLONG            Offset; 

    Offset = (LONGLONG) dwBlk;
    Offset = Offset * Vcb->BlockSize;

    if (!Ext2CopyRead(
            Vcb->StreamObj,
            (PLARGE_INTEGER)&Offset,
            Vcb->BlockSize,
            PIN_WAIT,
            Buffer,
            &IoStatus ));

    if (!NT_SUCCESS(IoStatus.Status)) {
        return FALSE;
    }

    return TRUE;
}


BOOLEAN
Ext2SaveBlock ( IN PEXT2_IRP_CONTEXT    IrpContext,
                IN PEXT2_VCB            Vcb,
                IN ULONG                dwBlk,
                IN PVOID                Buf )
{
    LONGLONG Offset;
    BOOLEAN  bRet;

    Offset = (LONGLONG) dwBlk;
    Offset = Offset * Vcb->BlockSize;

    bRet = Ext2SaveBuffer(IrpContext, Vcb, Offset, Vcb->BlockSize, Buf);

    if (IsFlagOn(Vcb->Flags, VCB_FLOPPY_DISK)) {
        Ext2StartFloppyFlushDpc(Vcb, NULL, NULL);
    }

    return bRet;
}

BOOLEAN
Ext2SaveBuffer( IN PEXT2_IRP_CONTEXT    IrpContext,
                IN PEXT2_VCB            Vcb,
                IN LONGLONG             Offset,
                IN ULONG                Size,
                IN PVOID                Buf )
{
    PBCB        Bcb;
    PVOID       Buffer;
    BOOLEAN     bRet;

    if( !CcPinRead( Vcb->StreamObj,
                    (PLARGE_INTEGER) (&Offset),
                    Size,
                    PIN_WAIT,
                    &Bcb,
                    &Buffer )) {

        Ext2Print((DBG_ERROR, "Ext2SaveBuffer: PinReading error ...\n"));
        return FALSE;
    }

    __try {

        RtlCopyMemory(Buffer, Buf, Size);
        CcSetDirtyPinnedData(Bcb, NULL );
        Ext2RepinBcb(IrpContext, Bcb);

        SetFlag(Vcb->StreamObj->Flags, FO_FILE_MODIFIED);

        Ext2AddMcbEntry(Vcb, Offset, (LONGLONG)Size);

        bRet = TRUE;

    } __finally {

        if (AbnormalTermination()) {

            CcUnpinData(Bcb);
            bRet = FALSE;
        }
    }

    CcUnpinData(Bcb);

    return bRet;
}

NTSTATUS
Ext2GetBlock(
    IN PEXT2_IRP_CONTEXT    IrpContext,
    IN PEXT2_VCB            Vcb,
    IN ULONG                dwContent,
    IN ULONG                Index,
    IN ULONG                Layer,
    IN BOOLEAN              bAlloc,
    OUT PULONG              pBlock
    )
{
    NTSTATUS    Status = STATUS_SUCCESS;
    ULONG       *pData = NULL;
    ULONG       i = 0, j = 0, temp = 1;
    ULONG       dwRet = 0;
    ULONG       dwBlk = 0;

    if (Layer == 0) {

        dwRet = dwContent;

    } else if (Layer <= 3) {

        /* allocate memory for pData to contain the block */
        pData = (ULONG *)ExAllocatePoolWithTag(
                        NonPagedPool,
                        Vcb->BlockSize,
                        EXT2_DATA_MAGIC
                    );
        if (!pData) {
            Ext2Print((DBG_ERROR, "Ext2GetBlock: no enough memory.\n"));
            Status = STATUS_INSUFFICIENT_RESOURCES;
            goto errorout;
        }

        /* load the block dwContext into pData buffer */
        if (!Ext2LoadBlock(Vcb, dwContent, pData)) {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            goto errorout;
        }

        temp = 1 << ((BLOCK_BITS - 2) * (Layer - 1));
        
        i = Index / temp;
        j = Index % temp;

        dwBlk = pData[i];

        if (dwBlk == 0) {

            if (bAlloc) {

                /* we need allocate new block: dwBlk */

                Status = Ext2NewBlock(
                            IrpContext,
                            Vcb,
                            0,
                            dwContent,
                            &dwBlk
                            );

                if (!NT_SUCCESS(Status)) {
                    goto errorout;
                }

                /* we need save pData now */
                pData[i] = dwBlk;

                /* save the block first, before next call */
                if (!Ext2SaveBlock( IrpContext, Vcb,
                                    dwContent, pData)) {
                    /* error occurs when saving the block */
                    Status = STATUS_INSUFFICIENT_RESOURCES;
                    /* we need free newly allocated block */
                    Ext2FreeBlock(IrpContext, Vcb, dwBlk);
                    goto errorout;
                }
            } else {
                goto errorout;
            }
        }

        /* transfer to next recursion call */
        Status = Ext2GetBlock(
                    IrpContext,
                    Vcb,
                    dwBlk,
                    j,
                    Layer - 1,
                    bAlloc,
                    &dwRet
                    );

        if (!NT_SUCCESS(Status)) {
            dwRet = 0;
        }
    }

errorout:

    /* free the memory of pData */
    if (pData) {
        ExFreePoolWithTag(pData, EXT2_DATA_MAGIC);
    }

    *pBlock = dwRet;

    return Status;
}

NTSTATUS
Ext2BlockMap(
    IN PEXT2_IRP_CONTEXT    IrpContext,
    IN PEXT2_VCB            Vcb,
    IN ULONG                InodeNo,
    IN PEXT2_INODE          Inode,
    IN ULONG                Index,
    IN BOOLEAN              bAlloc,
    OUT PULONG              pBlock
    )
{
    ULONG       i;

    NTSTATUS    Status = STATUS_SUCCESS;;

    *pBlock = 0;

    for (i = 0; i < EXT2_BLOCK_TYPES; i++) {

        if (Index < Vcb->NumBlocks[i]) {

            ULONG   dwRet = 0;
            ULONG   dwBlk = 0;

            dwBlk = Inode->i_block[i==0 ? (Index):(i + EXT2_NDIR_BLOCKS - 1)];

            if (dwBlk == 0) {

                if (!bAlloc) {

                    goto errorout;

                } else {

                    DbgBreak();

                    /* we need allocate new block: dwBlk */
                    Status = Ext2NewBlock(
                                IrpContext,
                                Vcb,
                                (InodeNo - 1) / BLOCKS_PER_GROUP,
                                0,
                                &dwBlk
                                );

                    if (!NT_SUCCESS(Status)) {
                        goto errorout;
                    }

                    /* save the it into inode*/
                    Inode->i_block[i==0 ? (Index):(i + EXT2_NDIR_BLOCKS - 1)] = dwBlk;

                    /* save the inode */
                    if (!Ext2SaveInode( IrpContext,
                                        Vcb,
                                        InodeNo,
                                        Inode)) {

                        Status = STATUS_UNSUCCESSFUL;

                        Ext2FreeBlock(IrpContext, Vcb, dwBlk);

                        goto errorout;
                    }
                }
            }

            Status = Ext2GetBlock(
                        IrpContext,
                        Vcb,
                        dwBlk,
                        Index,
                        i,
                        bAlloc,
                        &dwRet
                        );

            if (NT_SUCCESS(Status)) {
                *pBlock = dwRet;
            }

            break;
        }

        Index -= Vcb->NumBlocks[i];
    }

errorout:

    return Status;
}


NTSTATUS
Ext2BuildBDL( 
    IN PEXT2_IRP_CONTEXT    IrpContext,
    IN PEXT2_VCB            Vcb,
    IN ULONG                InodeNo,
    IN PEXT2_INODE          Inode,
    IN ULONGLONG            Offset, 
    IN ULONG                Size, 
    IN BOOLEAN              bAlloc,
    OUT PEXT2_BDL *         Bdls,
    OUT PULONG              Count
    )
{
    ULONG       nBeg, nEnd;
    ULONG       nBlocks;
    ULONG       dwBlk, i;
    ULONG       dwBytes = 0;
    LONGLONG    Lba = 0;
    NTSTATUS    Status = STATUS_SUCCESS;

    PEXT2_BDL   Bdl = NULL;

    *Bdls = NULL;

    if ((IrpContext && IrpContext->Irp) && 
        ((IrpContext->Irp->Flags & IRP_NOCACHE) || 
         (IrpContext->Irp->Flags & IRP_PAGING_IO))) {
        Size = (Size + SECTOR_SIZE - 1) & (~(SECTOR_SIZE - 1));
    }

    nBeg = (ULONG)(Offset >> BLOCK_BITS);
    nEnd = (ULONG)((Size + Offset + Vcb->BlockSize - 1) >> BLOCK_BITS);

    nBlocks = 0;

    Bdl = ExAllocatePoolWithTag(
                PagedPool,
                sizeof(EXT2_BDL) * (nEnd - nBeg),
                EXT2_BDL_MAGIC
            );

    if (!Bdl) {
        Ext2Print((DBG_ERROR, "Ex2BuildBDL: failed to allocate Bdl.\n"));
        Status = STATUS_INSUFFICIENT_RESOURCES;
        goto errorout;
    }

    RtlZeroMemory(Bdl, sizeof(EXT2_BDL) * (nEnd - nBeg));
            
    for (i = nBeg; i < nEnd; i++) {

        Status = Ext2BlockMap(
                    IrpContext,
                    Vcb,
                    InodeNo,
                    Inode,
                    i,
                    bAlloc,
                    &dwBlk
                    );

        if (!NT_SUCCESS(Status)) {

            DbgBreak();
            ExFreePoolWithTag(Bdl, EXT2_BLD_MAGIC);
            Bdl = NULL;
            goto errorout;
        }

        if (dwBlk > 0) { /* the block exist */

            Lba = ((LONGLONG) dwBlk) << BLOCK_BITS;
                
            if (nBeg == nEnd - 1) { /* only one block is to be read */

                dwBytes = Size;

                Bdl[nBlocks].Lba = Lba + (LONGLONG)(Offset & (Vcb->BlockSize - 1));
                Bdl[nBlocks].Length = dwBytes;
                Bdl[nBlocks].Offset = 0;

                nBlocks++;

            } else { /* several blocks cases */

                if (i == nBeg) {  /* the frist block */

                    dwBytes = Vcb->BlockSize - (ULONG)(Offset & (Vcb->BlockSize - 1));

                    Bdl[nBlocks].Lba = Lba + (LONGLONG)(Offset & (Vcb->BlockSize - 1));
                    Bdl[nBlocks].Length = dwBytes;
                    Bdl[nBlocks].Offset = 0;

                    nBlocks++;

                } else if (i == nEnd - 1) { /* the last block */

                    if ((Bdl[nBlocks - 1].Lba + Bdl[nBlocks - 1].Length == Lba) && 
                        (Bdl[nBlocks - 1].Offset + Bdl[nBlocks - 1].Length == dwBytes)) {
                        Bdl[nBlocks - 1].Length += Size - dwBytes;
                    } else {
                        Bdl[nBlocks].Lba = Lba;
                        Bdl[nBlocks].Length = Size - dwBytes;
                        Bdl[nBlocks].Offset = dwBytes;
                        nBlocks++;
                    }

                    dwBytes = Size;

                } else { /* sequential blocks */

                    if ((Bdl[nBlocks - 1].Lba + Bdl[nBlocks - 1].Length == Lba) && 
                        (Bdl[nBlocks - 1].Offset + Bdl[nBlocks - 1].Length == dwBytes)) {
                        Bdl[nBlocks - 1].Length += Vcb->BlockSize;
                    } else {
                        Bdl[nBlocks].Lba = Lba;
                        Bdl[nBlocks].Length = Vcb->BlockSize;
                        Bdl[nBlocks].Offset = dwBytes;
                        nBlocks++;
                    }

                    dwBytes +=  Vcb->BlockSize;
                }
            }

        } else {

            if (nBeg == nEnd - 1) {
                dwBytes = Size;
            } else {

                if (i == nBeg) {  /* the frist block */
                    dwBytes = Vcb->BlockSize - (ULONG)(Offset & (Vcb->BlockSize - 1));
                } else if (i == nEnd - 1) { /* the last block */
                    dwBytes = Size;
                } else { /* sequential blocks */
                    dwBytes +=  Vcb->BlockSize;
                }
            }
        }
    }

    *Bdls  = Bdl;
    *Count = nBlocks;

    INC_MEM_COUNT(10, Bdl);

errorout:

    return Status;
}

NTSTATUS
Ext2NewBlock(
    PEXT2_IRP_CONTEXT IrpContext,
    PEXT2_VCB Vcb,
    ULONG     GroupHint,
    ULONG     BlockHint,  
    PULONG    dwRet )
{
    RTL_BITMAP      BlockBitmap;
    LARGE_INTEGER   Offset;
    ULONG           Length = 0;

    PBCB            BitmapBcb;
    PVOID           BitmapCache;

    ULONG           Group = 0, dwBlk, dwHint = 0;

    *dwRet = 0;
    dwBlk = 0XFFFFFFFF;

    if (GroupHint > Vcb->NumOfGroups)
        GroupHint = Vcb->NumOfGroups - 1;

    if (BlockHint != 0) {
        GroupHint = (BlockHint - EXT2_FIRST_DATA_BLOCK) / BLOCKS_PER_GROUP;
        dwHint = (BlockHint - EXT2_FIRST_DATA_BLOCK) % BLOCKS_PER_GROUP;
    }
  
ScanBitmap:
  
    // Perform Prefered Group
    if (Vcb->GroupDesc[GroupHint].bg_free_blocks_count) {

        Offset.QuadPart = (LONGLONG) Vcb->BlockSize;
        Offset.QuadPart = Offset.QuadPart * 
                          Vcb->GroupDesc[GroupHint].bg_block_bitmap;

        if (GroupHint == Vcb->NumOfGroups - 1) {

            Length = TOTAL_BLOCKS % BLOCKS_PER_GROUP;

            /* s_blocks_count is integer multiple of s_blocks_per_group */
            if (Length == 0) {
                Length = BLOCKS_PER_GROUP;
            }
        } else {
            Length = BLOCKS_PER_GROUP;
        }

        if (!CcPinRead( Vcb->StreamObj,
                        &Offset,
                        Vcb->BlockSize,
                        PIN_WAIT,
                        &BitmapBcb,
                        &BitmapCache ) ) {

            Ext2Print((DBG_ERROR, "Ext2NewBlock: PinReading error ...\n"));
            return STATUS_INSUFFICIENT_RESOURCES;
        }

        RtlInitializeBitMap( &BlockBitmap,
                             BitmapCache,
                             Length );

        Group = GroupHint;

        if (RtlCheckBit(&BlockBitmap, dwHint) == 0) {
            dwBlk = dwHint;
        } else  {
            dwBlk = RtlFindClearBits(&BlockBitmap, 1, dwHint);
        }

        // We could not get new block in the prefered group.
        if (dwBlk == 0xFFFFFFFF) {

            CcUnpinData(BitmapBcb);
            BitmapBcb = NULL;
            BitmapCache = NULL;

            RtlZeroMemory(&BlockBitmap, sizeof(RTL_BITMAP));
        }
    }

    if (dwBlk == 0xFFFFFFFF) {

        for(Group = 0; Group < Vcb->NumOfGroups; Group++)
        if (Vcb->GroupDesc[Group].bg_free_blocks_count) {
            if (Group == GroupHint)
                continue;

            Offset.QuadPart = (LONGLONG) Vcb->BlockSize;
            Offset.QuadPart = Offset.QuadPart * Vcb->GroupDesc[Group].bg_block_bitmap;

            if (Vcb->NumOfGroups == 1) {
                Length = TOTAL_BLOCKS;
            } else {
                if (Group == Vcb->NumOfGroups - 1) {

                    Length = TOTAL_BLOCKS % BLOCKS_PER_GROUP;

                    /* s_blocks_count is integer multiple of s_blocks_per_group */
                    if (Length == 0) {
                        Length = BLOCKS_PER_GROUP;
                    }
                } else {
                    Length = BLOCKS_PER_GROUP;
                }
            }

            if (!CcPinRead( Vcb->StreamObj,
                            &Offset,
                            Vcb->BlockSize,
                            PIN_WAIT,
                            &BitmapBcb,
                            &BitmapCache ) ) {
                Ext2Print((DBG_ERROR, "Ext2NewBlock: PinReading error ...\n"));
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            RtlInitializeBitMap( &BlockBitmap,
                                 BitmapCache,
                                 Length );

            dwBlk = RtlFindClearBits(&BlockBitmap, 1, 0);

            if (dwBlk != 0xFFFFFFFF) {
                break;

            } else {

                CcUnpinData(BitmapBcb);
                BitmapBcb = NULL;
                BitmapCache = NULL;

                RtlZeroMemory(&BlockBitmap, sizeof(RTL_BITMAP));
            }
        }
    }
        
    if (dwBlk < Length) {

        RtlSetBits(&BlockBitmap, dwBlk, 1);

        CcSetDirtyPinnedData(BitmapBcb, NULL );

        Ext2RepinBcb(IrpContext, BitmapBcb);

        CcUnpinData(BitmapBcb);

        Ext2AddMcbEntry(Vcb, Offset.QuadPart, (LONGLONG)Vcb->BlockSize);

        *dwRet = dwBlk + EXT2_FIRST_DATA_BLOCK + Group * BLOCKS_PER_GROUP;

        //Updating Group Desc / Superblock
        Vcb->GroupDesc[Group].bg_free_blocks_count--;
        Ext2SaveGroup(IrpContext, Vcb, Group);

        Vcb->SuperBlock->s_free_blocks_count--;
        Ext2SaveSuper(IrpContext, Vcb);

        {
            ULONG i=0;
            for (i=0; i < Vcb->NumOfGroups; i++)
            {
                if ((Vcb->GroupDesc[i].bg_block_bitmap == *dwRet) ||
                    (Vcb->GroupDesc[i].bg_inode_bitmap == *dwRet) ||
                    (Vcb->GroupDesc[i].bg_inode_table == *dwRet) ) {
                    DbgBreak();
                    GroupHint = Group;
                    goto ScanBitmap;
                }
            }
        }

        return STATUS_SUCCESS;
    }

    return STATUS_DISK_FULL;
}

NTSTATUS
Ext2FreeBlock(
    PEXT2_IRP_CONTEXT IrpContext,
    PEXT2_VCB Vcb,
    ULONG     Block )
{
    RTL_BITMAP      BlockBitmap;
    LARGE_INTEGER   Offset;
    ULONG           Length;

    PBCB            BitmapBcb;
    PVOID           BitmapCache;

    ULONG           Group, dwBlk;
    BOOLEAN         bModified = FALSE;

    if ( Block < EXT2_FIRST_DATA_BLOCK || 
         (Block / BLOCKS_PER_GROUP) >= Vcb->NumOfGroups) {

        DbgBreak();
        return STATUS_INVALID_PARAMETER;
    }

    Ext2Print((DBG_INFO, "Ext2FreeBlock: Block %xh to be freed.\n", Block));

    Group = (Block - EXT2_FIRST_DATA_BLOCK) / BLOCKS_PER_GROUP;

    dwBlk = (Block - EXT2_FIRST_DATA_BLOCK) % BLOCKS_PER_GROUP;
    
    {
        Offset.QuadPart = (LONGLONG) Vcb->BlockSize;
        Offset.QuadPart = Offset.QuadPart * Vcb->GroupDesc[Group].bg_block_bitmap;

        if (Group == Vcb->NumOfGroups - 1) {

            Length = TOTAL_BLOCKS % BLOCKS_PER_GROUP;

            /* s_blocks_count is integer multiple of s_blocks_per_group */
            if (Length == 0) {
                Length = BLOCKS_PER_GROUP;
            }

        } else {
            Length = BLOCKS_PER_GROUP;
        }

        if (!CcPinRead( Vcb->StreamObj,
                        &Offset,
                        Vcb->BlockSize,
                        PIN_WAIT,
                        &BitmapBcb,
                        &BitmapCache ) ) {

            Ext2Print((DBG_ERROR, "Ext2DeleteBlock: PinReading error ...\n"));
            return STATUS_INSUFFICIENT_RESOURCES;
        }

        RtlInitializeBitMap( &BlockBitmap,
                             BitmapCache,
                             Length );

        if (RtlCheckBit(&BlockBitmap, dwBlk) == 0) {
            
        } else {
            RtlClearBits(&BlockBitmap, dwBlk, 1);
            bModified = TRUE;
        }

        if (!bModified) {

            CcUnpinData(BitmapBcb);
            BitmapBcb = NULL;
            BitmapCache = NULL;

            RtlZeroMemory(&BlockBitmap, sizeof(RTL_BITMAP));
        }
    }
        
    if (bModified) {

        CcSetDirtyPinnedData(BitmapBcb, NULL );

        Ext2RepinBcb(IrpContext, BitmapBcb);

        CcUnpinData(BitmapBcb);

        Ext2AddMcbEntry(Vcb, Offset.QuadPart, (LONGLONG)Vcb->BlockSize);

        //Updating Group Desc / Superblock
        Vcb->GroupDesc[Group].bg_free_blocks_count++;
        Ext2SaveGroup(IrpContext, Vcb, Group);

        Vcb->SuperBlock->s_free_blocks_count++;
        Ext2SaveSuper(IrpContext, Vcb);

        return STATUS_SUCCESS;
    }

    return STATUS_UNSUCCESSFUL;
}

NTSTATUS
Ext2ExpandBlock(
    PEXT2_IRP_CONTEXT IrpContext,
    PEXT2_VCB   Vcb,
    PEXT2_FCB   Fcb,
    ULONG   dwContent,
    ULONG   Index,
    ULONG   layer,
    BOOLEAN bNew,
    ULONG   *dwRet )
{
    ULONG       *pData = NULL;
    ULONG       i = 0, j = 0, temp = 1;
    ULONG       dwNewBlk = 0, dwBlk = 0;
    BOOLEAN     bDirty = FALSE;
    NTSTATUS    Status = STATUS_SUCCESS;

    PEXT2_INODE       Inode  = Fcb->Inode;
    PEXT2_SUPER_BLOCK Ext2Sb = Vcb->SuperBlock;

    pData = (ULONG *) ExAllocatePoolWithTag(
                PagedPool,
                Vcb->BlockSize,
                EXT2_DATA_MAGIC
                );

    if (!pData) {
        Ext2Print((DBG_ERROR, "Ex2ExpandBlock: failed to allocate pData.\n"));
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    RtlZeroMemory(pData, Vcb->BlockSize);

    if (bNew) {

        if (layer == 0) {

            if (IsDirectory(Fcb)) {

                PEXT2_DIR_ENTRY2 pEntry;

                pEntry = (PEXT2_DIR_ENTRY2) pData;
                pEntry->rec_len = (USHORT)(Vcb->BlockSize);

                if (!Ext2SaveBlock(IrpContext, Vcb, dwContent, (PVOID)pData)) {

                    Status = STATUS_UNSUCCESSFUL;
                    goto errorout;
                }

            } else {

                LARGE_INTEGER   Offset;
            
                Offset.QuadPart  = (LONGLONG) dwContent;
                Offset.QuadPart = Offset.QuadPart * Vcb->BlockSize;

                Ext2RemoveMcbEntry(Vcb, Offset.QuadPart, (LONGLONG)Vcb->BlockSize);
            }
        } else {

            if (!Ext2SaveBlock(IrpContext, Vcb, dwContent, (PVOID)pData)) {

                Status = STATUS_UNSUCCESSFUL;
                goto errorout;
            }
        }
    }

    if (layer == 0) {

        dwNewBlk = dwContent;

    } else if (layer <= 3) {

        if (!bNew) {

            if (!Ext2LoadBlock(Vcb, dwContent, (void *)pData)) {

                Status = STATUS_UNSUCCESSFUL;
                goto errorout;
            }
        }

        temp = 1 << ((BLOCK_BITS - 2) * (layer - 1));

        i = Index / temp;
        j = Index % temp;

        dwBlk = pData[i];

        if (dwBlk == 0) {

            Status = Ext2NewBlock(
                        IrpContext,
                        Vcb, 0,
                        dwContent,
                       &dwBlk);

            if (!NT_SUCCESS(Status)) {

                Ext2Print((DBG_ERROR, "Ext2ExpandBlock: get new block error.\n"));
                goto errorout;
            }

            Inode->i_blocks += (Vcb->BlockSize / SECTOR_SIZE);

            pData[i] = dwBlk;
            bDirty = TRUE;
        }

        Status = Ext2ExpandBlock(
                    IrpContext,
                    Vcb, Fcb,
                    dwBlk, j,
                    layer - 1,
                    bDirty,
                    &dwNewBlk );

        if (!NT_SUCCESS(Status))
        {
            Ext2Print((DBG_ERROR, "Ext2ExpandBlockk: ... error recuise...\n"));
            goto errorout;
        }
        
        if (bDirty)
        {
            Ext2SaveBlock( IrpContext,
                            Vcb, dwContent,
                            (void *)pData );
        }
    }

errorout:

    if (pData) {
        ExFreePoolWithTag(pData, EXT2_DATA_MAGIC);
    }

    if (NT_SUCCESS(Status) && dwRet) {
        *dwRet = dwNewBlk;
    }

    return Status;
}


NTSTATUS
Ext2ExpandInode(
    IN PEXT2_IRP_CONTEXT    IrpContext,
    IN PEXT2_VCB            Vcb,
    IN PEXT2_FCB            Fcb,
    OUT PULONG              dwRet
    )
{
    ULONG Index = 0;
    ULONG dwBlk = 0, dwNewBlk = 0;
    ULONG    i;
    NTSTATUS Status = STATUS_SUCCESS;
    BOOLEAN  bNewBlock = FALSE;

    PEXT2_INODE Inode = Fcb->Inode;

    Index = (ULONG)(Fcb->Header.AllocationSize.QuadPart >> BLOCK_BITS);

    if (Index >= Vcb->MaxBlocks) {
        Ext2Print((DBG_ERROR, "Ext2ExpandInode: beyond the maxinum size of an inode.\n"));
        return STATUS_UNSUCCESSFUL;
    }

    for (i = 0; i < EXT2_BLOCK_TYPES; i++) {

        if (Index < Vcb->NumBlocks[i]) {

            dwBlk = Inode->i_block[i==0 ? (Index):(i + EXT2_NDIR_BLOCKS - 1)];

            if (dwBlk == 0) {

                Status = Ext2NewBlock(
                            IrpContext,
                            Vcb,
                            Fcb->BlkHint ? 0 : ((Fcb->Mcb->iNo - 1) / INODES_PER_GROUP),
                            Fcb->BlkHint,
                            &dwBlk );

                if (!NT_SUCCESS(Status) ) {
                    Ext2Print((DBG_ERROR, "Ext2ExpandInode: get new block error.\n"));
                    break;
                }

                Inode->i_block[i==0 ? (Index):(i + EXT2_NDIR_BLOCKS - 1)] = dwBlk;
                Inode->i_blocks += (Vcb->BlockSize / SECTOR_SIZE);
                bNewBlock = TRUE;
            }

            Status = Ext2ExpandBlock (
                            IrpContext,
                            Vcb, Fcb,
                            dwBlk, Index,
                            i, bNewBlock,
                            &dwNewBlk  ); 

            if (NT_SUCCESS(Status)) {
                Fcb->Header.AllocationSize.QuadPart += (LONGLONG)Vcb->BlockSize;
            }

            break;
        }

        Index -= Vcb->NumBlocks[i];
    }

    Ext2SaveInode(IrpContext, Vcb, Fcb->Mcb->iNo, Inode);

    if (NT_SUCCESS(Status)) {
        if (dwNewBlk) {
            if (dwRet) {
                Fcb->BlkHint = dwNewBlk+1;
                *dwRet = dwNewBlk;

                Ext2Print((DBG_INFO, "Ext2ExpandInode: %S (%xh) i=%2.2xh Index=%8.8xh New Block=%8.8xh\n",
                           Fcb->Mcb->ShortName.Buffer, Fcb->Mcb->iNo, i, Index, dwNewBlk));
            }
        } else {
            DbgBreak();
            Status = STATUS_UNSUCCESSFUL;
        }
    }

    return Status;
}

NTSTATUS
Ext2NewInode(
            PEXT2_IRP_CONTEXT IrpContext,
            PEXT2_VCB Vcb,
            ULONG   GroupHint,
            ULONG   Type,
            PULONG  Inode )
{
    RTL_BITMAP      InodeBitmap;
    PVOID           BitmapCache;
    PBCB            BitmapBcb;

    ULONG           Group, i, j;
    ULONG           Average, Length;
    LARGE_INTEGER   Offset;
    
    ULONG           dwInode;

    *Inode = dwInode = 0XFFFFFFFF;

    ExAcquireResourceExclusiveLite(&Vcb->MetaResource, TRUE);

repeat:

    Group = i = 0;
    
    if (Type == EXT2_FT_DIR) {

        Average = Vcb->SuperBlock->s_free_inodes_count / Vcb->NumOfGroups;

        for (j = 0; j < Vcb->NumOfGroups; j++) {

            i = (j + GroupHint) % (Vcb->NumOfGroups);

            if ((Vcb->GroupDesc[i].bg_used_dirs_count << 8) < 
                 Vcb->GroupDesc[i].bg_free_inodes_count ) {
                Group = i + 1;
                break;
            }
        }

        if (!Group) {

            for (j = 0; j < Vcb->NumOfGroups; j++) {

                if (Vcb->GroupDesc[j].bg_free_inodes_count >= Average) {
                    if (!Group || (Vcb->GroupDesc[j].bg_free_blocks_count >
                                   Vcb->GroupDesc[Group].bg_free_blocks_count ))
                        Group = j + 1;
                }
            }
        }

    } else {

        /*
         * Try to place the inode in its parent directory (GroupHint)
         */

        if (Vcb->GroupDesc[GroupHint].bg_free_inodes_count) {

            Group = GroupHint + 1;

        } else {

            i = GroupHint;

            /*
             * Use a quadratic hash to find a group with a
             * free inode
             */

            for (j = 1; j < Vcb->NumOfGroups; j <<= 1) {

                i += j;
                if (i > Vcb->NumOfGroups) 
                    i -= Vcb->NumOfGroups;

                if (Vcb->GroupDesc[i].bg_free_inodes_count) {
                    Group = i + 1;
                    break;
                }
            }
        }

        if (!Group) {
            /*
             * That failed: try linear search for a free inode
             */
            i = GroupHint + 1;
            for (j = 2; j < Vcb->NumOfGroups; j++) {
                if (++i >= Vcb->NumOfGroups) i = 0;

                if (Vcb->GroupDesc[i].bg_free_inodes_count) {
                    Group = i + 1;
                    break;
                }
            }
        }
    }

    // Could not find a proper group.
    if (!Group) {
        ExReleaseResourceLite(&Vcb->MetaResource);
        return STATUS_DISK_FULL;
    }

    {
        Group--;

        Offset.QuadPart = (LONGLONG) Vcb->BlockSize;
        Offset.QuadPart = Offset.QuadPart * Vcb->GroupDesc[Group].bg_inode_bitmap;

        if (Vcb->NumOfGroups == 1) {
            Length = INODES_COUNT;
        } else {
            if (Group == Vcb->NumOfGroups - 1) {
                Length = INODES_COUNT % INODES_PER_GROUP;
                if (!Length) {
                    /* INODES_COUNT is integer multiple of INODES_PER_GROUP */
                    Length = INODES_PER_GROUP;
                }
            } else  {
                Length = INODES_PER_GROUP;
            }
        }
        
        if (!CcPinRead( Vcb->StreamObj,
                        &Offset,
                        Vcb->BlockSize,
                        PIN_WAIT,
                        &BitmapBcb,
                        &BitmapCache ) ) {

            Ext2Print((DBG_ERROR, "Ext2NewInode: PinReading error ...\n"));
            ExReleaseResourceLite(&Vcb->MetaResource);
            return STATUS_UNSUCCESSFUL;
        }

        RtlInitializeBitMap( &InodeBitmap,
                             BitmapCache,
                             Length );

        dwInode = RtlFindClearBits(&InodeBitmap, 1, 0);

        if (dwInode == 0xFFFFFFFF) {
            CcUnpinData(BitmapBcb);
            BitmapBcb = NULL;
            BitmapCache = NULL;

            RtlZeroMemory(&InodeBitmap, sizeof(RTL_BITMAP));
        }
    }

    if (dwInode == 0xFFFFFFFF || dwInode >= Length) {

        if (Vcb->GroupDesc[Group].bg_free_inodes_count != 0) {

            Vcb->GroupDesc[Group].bg_free_inodes_count = 0;

            Ext2SaveGroup(IrpContext, Vcb, Group);
        }

        goto repeat;

    } else {

        RtlSetBits(&InodeBitmap, dwInode, 1);

        CcSetDirtyPinnedData(BitmapBcb, NULL );

        Ext2RepinBcb(IrpContext, BitmapBcb);

        CcUnpinData(BitmapBcb);

        Ext2AddMcbEntry(Vcb, Offset.QuadPart, (LONGLONG)Vcb->BlockSize);

        *Inode = dwInode + 1 + Group * INODES_PER_GROUP;

        //Updating Group Desc / Superblock
        Vcb->GroupDesc[Group].bg_free_inodes_count--;
        if (Type == EXT2_FT_DIR) {
            Vcb->GroupDesc[Group].bg_used_dirs_count++;
        }

        Ext2SaveGroup(IrpContext, Vcb, Group);

        Vcb->SuperBlock->s_free_inodes_count--;
        Ext2SaveSuper(IrpContext, Vcb);
        ExReleaseResourceLite(&Vcb->MetaResource);
        
        return STATUS_SUCCESS;        
    }

    ExReleaseResourceLite(&Vcb->MetaResource);
    return STATUS_DISK_FULL;
}

BOOLEAN
Ext2FreeInode(
            PEXT2_IRP_CONTEXT IrpContext,
            PEXT2_VCB Vcb,
            ULONG Inode,
            ULONG Type )
{
    RTL_BITMAP      InodeBitmap;
    PVOID           BitmapCache;
    PBCB            BitmapBcb;

    ULONG           Group;
    ULONG           Length;
    LARGE_INTEGER   Offset;

    ULONG           dwIno;
    BOOLEAN         bModified = FALSE;
 
    ExAcquireResourceExclusiveLite(&Vcb->MetaResource, TRUE);   

    Group = (Inode - 1) / INODES_PER_GROUP;
    dwIno = (Inode - 1) % INODES_PER_GROUP;

    Ext2Print((DBG_INFO, "Ext2FreeInode: Inode: %xh (Group/Off = %xh/%xh)\n",
                         Inode, Group, dwIno));
    
    {
        Offset.QuadPart = (LONGLONG) Vcb->BlockSize;
        Offset.QuadPart = Offset.QuadPart * Vcb->GroupDesc[Group].bg_inode_bitmap;
        if (Group == Vcb->NumOfGroups - 1) {

            Length = INODES_COUNT % INODES_PER_GROUP;
            if (!Length) {
                /* s_inodes_count is integer multiple of s_inodes_per_group */
                Length = INODES_PER_GROUP;
            }
        } else {
            Length = INODES_PER_GROUP;
        }

        if (!CcPinRead( Vcb->StreamObj,
                        &Offset,
                        Vcb->BlockSize,
                        PIN_WAIT,
                        &BitmapBcb,
                        &BitmapCache ) ) {
            Ext2Print((DBG_ERROR, "Ext2FreeInode: PinReading error ...\n"));
            ExReleaseResourceLite(&Vcb->MetaResource);
            return FALSE;
        }

        RtlInitializeBitMap( &InodeBitmap,
                             BitmapCache,
                             Length );

        if (RtlCheckBit(&InodeBitmap, dwIno) == 0) {
            DbgBreak();
        } else {
            RtlClearBits(&InodeBitmap, dwIno, 1);
            bModified = TRUE;
        }

        if (!bModified) {

            CcUnpinData(BitmapBcb);
            BitmapBcb = NULL;
            BitmapCache = NULL;

            RtlZeroMemory(&InodeBitmap, sizeof(RTL_BITMAP));
        }
    }
        
    if (bModified) {

        CcSetDirtyPinnedData(BitmapBcb, NULL );

        Ext2RepinBcb(IrpContext, BitmapBcb);

        CcUnpinData(BitmapBcb);

        Ext2AddMcbEntry(Vcb, Offset.QuadPart, (LONGLONG)Vcb->BlockSize);

        //Updating Group Desc / Superblock
        if (Type == EXT2_FT_DIR) {
            Vcb->GroupDesc[Group].bg_used_dirs_count--;
        }

        Vcb->GroupDesc[Group].bg_free_inodes_count++;
        Ext2SaveGroup(IrpContext, Vcb, Group);

        Vcb->SuperBlock->s_free_inodes_count++;
        Ext2SaveSuper(IrpContext, Vcb);

        ExReleaseResourceLite(&Vcb->MetaResource);
        return TRUE;
    }

    ExReleaseResourceLite(&Vcb->MetaResource);
    return FALSE;
}


NTSTATUS
Ext2AddEntry (
         IN PEXT2_IRP_CONTEXT   IrpContext,
         IN PEXT2_VCB           Vcb,
         IN PEXT2_FCB           Dcb,
         IN ULONG               FileType,
         IN ULONG               Inode,
         IN PUNICODE_STRING     FileName )
{
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;

    PEXT2_DIR_ENTRY2        pDir = NULL;
    PEXT2_DIR_ENTRY2        pNewDir = NULL;
    PEXT2_DIR_ENTRY2        pTarget = NULL;

    ULONG                   Length = 0;
    ULONG                   dwBytes = 0;

    BOOLEAN                 bFound = FALSE;
    BOOLEAN                 bAdding = FALSE;

    BOOLEAN                 MainResourceAcquired = FALSE;

    ULONG                   dwRet = 0;

    if (!IsDirectory(Dcb)) {
        DbgBreak();
        return STATUS_NOT_A_DIRECTORY;
    }

    MainResourceAcquired = ExAcquireResourceExclusiveLite(&Dcb->MainResource, TRUE);

    __try {

        Dcb->ReferenceCount++;

        pDir = (PEXT2_DIR_ENTRY2) 
                    ExAllocatePoolWithTag(
                        PagedPool,
                        EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                        EXT2_DENTRY_MAGIC
                    );
        if (!pDir) {
            Ext2Print((DBG_ERROR, "Ex2AddEntry: failed to allocate pDir.\n"));
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        pTarget = (PEXT2_DIR_ENTRY2)
                    ExAllocatePoolWithTag(
                        PagedPool,
                        2 * EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                        EXT2_DENTRY_MAGIC
                    );
        if (!pTarget) {
            Ext2Print((DBG_ERROR, "Ex2AddEntry: failed to allocate pTarget.\n"));
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        if (IsFlagOn( SUPER_BLOCK->s_feature_incompat, 
                      EXT2_FEATURE_INCOMPAT_FILETYPE)) {
            pDir->file_type = (UCHAR) FileType;
        } else {
            pDir->file_type = 0;
        }

        {
            OEM_STRING  OemName;
            OemName.Buffer = pDir->name;
            OemName.MaximumLength = EXT2_NAME_LEN;
            OemName.Length  = 0;

            Status = Ext2UnicodeToOEM(Vcb, &OemName, FileName);

            if (!NT_SUCCESS(Status)) {
                __leave;
            }

            pDir->name_len = (CCHAR) OemName.Length;
        }

        pDir->inode  = Inode;
        pDir->rec_len = (USHORT) (EXT2_DIR_REC_LEN(pDir->name_len));

        dwBytes = 0;

Repeat:

        while ((LONGLONG)dwBytes < Dcb->Header.AllocationSize.QuadPart) {

            RtlZeroMemory(pTarget, EXT2_DIR_REC_LEN(EXT2_NAME_LEN));

            // Reading the DCB contents
            Status = Ext2ReadInode(
                        NULL,
                        Vcb,
                        Dcb->Mcb,
                        dwBytes,
                        (PVOID)pTarget,
                        EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                        &dwRet);

            if (!NT_SUCCESS(Status)) {
                Ext2Print((DBG_ERROR, "Ext2AddDirectory: Reading Directory Content error.\n"));
                __leave;
            }


            if ((pTarget->inode == 0) && (pTarget->rec_len == 0)) {
                if (Dcb->Header.AllocationSize.QuadPart > 
                   (LONGLONG) Vcb->BlockSize + dwBytes ) {
                    pTarget->rec_len = (__u16)(Vcb->BlockSize - dwBytes);
                } else {
                    pTarget->rec_len = (__u16)(Dcb->Header.AllocationSize.LowPart
                                               - dwBytes);
                }
            }

            if (((pTarget->inode == 0) && pTarget->rec_len >= pDir->rec_len) || 
                (pTarget->rec_len >= EXT2_DIR_REC_LEN(pTarget->name_len) + pDir->rec_len)) {

                if (pTarget->inode) {

                    RtlZeroMemory(pTarget, 2 * EXT2_DIR_REC_LEN(EXT2_NAME_LEN));

                    // Reading the DCB contents
                    Status = Ext2ReadInode(
                                NULL,
                                Vcb,
                                Dcb->Mcb,
                                dwBytes,
                                (PVOID)pTarget,
                                2 * EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                                &dwRet);

                    if (!NT_SUCCESS(Status)) {
                        Ext2Print((DBG_ERROR, "Ext2AddDirectory: Reading Directory Content error.\n"));
                        __leave;
                    }

                    Length = EXT2_DIR_REC_LEN(pTarget->name_len);
                    
                    pNewDir = (PEXT2_DIR_ENTRY2) ((PUCHAR)pTarget + EXT2_DIR_REC_LEN(pTarget->name_len));

                    pNewDir->rec_len = pTarget->rec_len - EXT2_DIR_REC_LEN(pTarget->name_len);

                    pTarget->rec_len = EXT2_DIR_REC_LEN(pTarget->name_len);

                } else {

                    Length  = 0;
                    pNewDir = pTarget;
                }

                pNewDir->file_type = pDir->file_type;
                pNewDir->inode = pDir->inode;
                pNewDir->name_len = pDir->name_len;
                memcpy(pNewDir->name, pDir->name, pDir->name_len);
                Length += EXT2_DIR_REC_LEN(pDir->name_len);

                bFound = TRUE;
                break;
            }
            
            dwBytes += pTarget->rec_len;
        }

        if (bFound) {

            if ( FileType==EXT2_FT_DIR ) {

                if(((pDir->name_len == 1) && (pDir->name[0] == '.')) ||
                   ((pDir->name_len == 2) && (pDir->name[0] == '.') && (pDir->name[1] == '.')) ) {
                } else {
                    Dcb->Inode->i_links_count++;
                }
            }

            Status = Ext2WriteInode(
                        IrpContext,
                        Vcb,
                        Dcb->Mcb,
                        (ULONGLONG)dwBytes,
                        pTarget,
                        Length,
                        FALSE,
                        &dwRet );
        } else {

            // We should expand the size of the dir inode 
            if (!bAdding) {

                Status = Ext2ExpandInode(IrpContext, Vcb, Dcb, &dwRet);

                if (NT_SUCCESS(Status)) {

                    Dcb->Inode->i_size = Dcb->Header.AllocationSize.LowPart;

                    Ext2SaveInode(IrpContext, Vcb, Dcb->Mcb->iNo, Dcb->Inode);

                    Dcb->Header.ValidDataLength = Dcb->Header.FileSize = 
                    Dcb->Mcb->FileSize = Dcb->Header.AllocationSize;

                    bAdding = TRUE;

                    goto Repeat;
                }

                __leave;

            } else { // Something must be error!
                __leave;
            }
        }

    } __finally {

        Dcb->ReferenceCount--;

        if(MainResourceAcquired)    {
            ExReleaseResourceLite(&Dcb->MainResource);
        }

        if (pTarget != NULL) {
            ExFreePoolWithTag(pTarget, EXT2_DENTRY_MAGIC);
        }

        if (pDir) {
            ExFreePoolWithTag(pDir, EXT2_DENTRY_MAGIC);
        }
    }
    
    return Status;
}


NTSTATUS
Ext2RemoveEntry (
         IN PEXT2_IRP_CONTEXT   IrpContext,
         IN PEXT2_VCB           Vcb,
         IN PEXT2_FCB           Dcb,
         IN ULONG               FileType,
         IN ULONG               Inode )
{
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;

    PEXT2_DIR_ENTRY2        pTarget = NULL;
    PEXT2_DIR_ENTRY2        pPrevDir = NULL;

    USHORT                  PrevRecLen = 0;

    ULONG                   Length = 0;
    ULONG                   dwBytes = 0;

    BOOLEAN                 MainResourceAcquired = FALSE;

    ULONG                   dwRet;

    if (!IsDirectory(Dcb)) {
        return STATUS_NOT_A_DIRECTORY;
    }

    MainResourceAcquired = 
            ExAcquireResourceExclusiveLite(&Dcb->MainResource, TRUE);

    __try {

        Dcb->ReferenceCount++;

        pTarget = (PEXT2_DIR_ENTRY2)
                        ExAllocatePoolWithTag( PagedPool,
                            EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                            EXT2_DENTRY_MAGIC
                        );
        if (!pTarget) {
            Ext2Print((DBG_ERROR, "Ex2RemoveEntry: failed to allocate pTarget.\n"));
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }
        
        pPrevDir = (PEXT2_DIR_ENTRY2)
                        ExAllocatePoolWithTag(
                            PagedPool,
                            EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                            EXT2_DENTRY_MAGIC
                        );
        if (!pPrevDir) {
            Ext2Print((DBG_ERROR, "Ex2RemoveEntry: failed to allocate pPrevDir.\n"));
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        dwBytes = 0;

        while ((LONGLONG)dwBytes < Dcb->Header.AllocationSize.QuadPart) {

            RtlZeroMemory(pTarget, EXT2_DIR_REC_LEN(EXT2_NAME_LEN));

            Status = Ext2ReadInode(
                        NULL,
                        Vcb,
                        Dcb->Mcb,
                        dwBytes,
                        (PVOID)pTarget,
                        EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                        &dwRet);

            if (!NT_SUCCESS(Status)) {

                Ext2Print((DBG_ERROR, "Ext2RemoveEntry: Reading Directory Content error.\n"));
                __leave;
            }

            //
            // Find it ! Then remove the entry from Dcb ...
            //

            if (pTarget->inode == Inode) {

                ULONG   RecLen;

                BOOLEAN bAcross = FALSE;

                if (((ULONG)PrevRecLen + pTarget->rec_len) > BLOCK_SIZE) {
                    bAcross = TRUE;
                } else {

                    dwRet = (dwBytes - PrevRecLen) & (~(BLOCK_SIZE - 1));
                    if (dwRet != (dwBytes & (~(BLOCK_SIZE - 1))))
                    {
                        bAcross = TRUE;
                    }
                }

                if (!bAcross) {

                    pPrevDir->rec_len += pTarget->rec_len;

                    RecLen = EXT2_DIR_REC_LEN(pTarget->name_len);

                    RtlZeroMemory(pTarget, RecLen);

                    Status = Ext2WriteInode(
                                IrpContext,
                                Vcb,
                                Dcb->Mcb,
                                dwBytes - PrevRecLen,
                                pPrevDir,
                                8,
                                FALSE,
                                &dwRet
                                );

                    ASSERT(NT_SUCCESS(Status));

                    Status = Ext2WriteInode(
                                IrpContext,
                                Vcb,
                                Dcb->Mcb,
                                dwBytes,
                                pTarget,
                                RecLen,
                                FALSE,
                                &dwRet
                                );

                    ASSERT(NT_SUCCESS(Status));

                } else {

                    RecLen = (ULONG)pTarget->rec_len;
                    if (RecLen > EXT2_DIR_REC_LEN(EXT2_NAME_LEN)) {

                        RtlZeroMemory(pTarget, EXT2_DIR_REC_LEN(EXT2_NAME_LEN));

                        pTarget->rec_len = (USHORT)RecLen;

                        Status = Ext2WriteInode(
                                    IrpContext,
                                    Vcb,
                                    Dcb->Mcb,
                                    dwBytes,
                                    pTarget,
                                    EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                                    FALSE,
                                    &dwRet
                                    );

                        ASSERT(NT_SUCCESS(Status));

                    } else {

                        RtlZeroMemory(pTarget, RecLen);
                        pTarget->rec_len = (USHORT)RecLen;

                        Status = Ext2WriteInode(
                                    IrpContext,
                                    Vcb,
                                    Dcb->Mcb,
                                    dwBytes,
                                    pTarget,
                                    RecLen,
                                    FALSE,
                                    &dwRet
                                    );

                        ASSERT(NT_SUCCESS(Status));
                    }
                }

                //
                // Error if it's the entry of dot or dot-dot or drop the parent's refer link
                //

                if (FileType == EXT2_FT_DIR) {

                    if(((pTarget->name_len == 1) && (pTarget->name[0] == '.')) ||
                       ((pTarget->name_len == 2) && (pTarget->name[0] == '.') && (pTarget->name[1] == '.')) ) {

                        DbgBreak();
                    } else {
                        Dcb->Inode->i_links_count--;
                    }
                }

                //
                // Update at least mtime/atime if !EXT2_FT_DIR.
                //

                if ( !Ext2SaveInode(
                        IrpContext,
                        Vcb,
                        Dcb->Mcb->iNo,
                        Dcb->Inode
                       ) ) {
                    Status = STATUS_UNSUCCESSFUL;
                }
              
                break;

            } else {

                RtlCopyMemory(pPrevDir, pTarget, EXT2_DIR_REC_LEN(EXT2_NAME_LEN));
                PrevRecLen = pTarget->rec_len;
            }

            dwBytes += pTarget->rec_len;
        }

    } __finally {

        Dcb->ReferenceCount--;

        if(MainResourceAcquired)
            ExReleaseResourceLite(&Dcb->MainResource);

        if (pTarget != NULL) {
            ExFreePoolWithTag(pTarget, EXT2_DENTRY_MAGIC);
        }

        if (pPrevDir != NULL) {
            ExFreePoolWithTag(pPrevDir, EXT2_DENTRY_MAGIC);
        }
    }
    
    return Status;
}

NTSTATUS
Ext2SetParentEntry (
         IN PEXT2_IRP_CONTEXT   IrpContext,
         IN PEXT2_VCB           Vcb,
         IN PEXT2_FCB           Dcb,
         IN ULONG               OldParent,
         IN ULONG               NewParent )
{
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;

    PEXT2_DIR_ENTRY2        pSelf   = NULL;
    PEXT2_DIR_ENTRY2        pParent = NULL;

    ULONG                   dwBytes = 0;

    BOOLEAN                 MainResourceAcquired = FALSE;

    ULONG                   Offset = 0;

    if (!IsDirectory(Dcb)) {
        return STATUS_NOT_A_DIRECTORY;
    }

    MainResourceAcquired = 
        ExAcquireResourceExclusiveLite(&Dcb->MainResource, TRUE);

    __try {

        Dcb->ReferenceCount++;

        pSelf = (PEXT2_DIR_ENTRY2)
                    ExAllocatePoolWithTag(
                        PagedPool,
                        EXT2_DIR_REC_LEN(1) + EXT2_DIR_REC_LEN(2),
                        EXT2_DENTRY_MAGIC
                    );
        if (!pSelf) {
            Ext2Print((DBG_ERROR, "Ex2SetParentEntry: failed to allocate pSelf.\n"));
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        dwBytes = 0;

        //
        // Reading the DCB contents
        //

        Status = Ext2ReadInode(
                        NULL,
                        Vcb,
                        Dcb->Mcb,
                        Offset,
                        (PVOID)pSelf,
                        EXT2_DIR_REC_LEN(1) + EXT2_DIR_REC_LEN(2),
                        &dwBytes );

        if (!NT_SUCCESS(Status)) {
            Ext2Print((DBG_ERROR, "Ext2SetParentEntry: Reading Directory Content error.\n"));
            __leave;
        }

        ASSERT(dwBytes == EXT2_DIR_REC_LEN(1) + EXT2_DIR_REC_LEN(2));

        pParent = (PEXT2_DIR_ENTRY2)((PUCHAR)pSelf + pSelf->rec_len);

        if (pParent->inode != OldParent) {
            DbgBreak();
        }

        pParent->inode = NewParent;

        Status = Ext2WriteInode(
                        IrpContext,
                        Vcb, 
                        Dcb->Mcb,
                        Offset,
                        pSelf,
                        dwBytes,
                        FALSE,
                        &dwBytes );

    } __finally {

        Dcb->ReferenceCount--;

        if(MainResourceAcquired)    {
            ExReleaseResourceLite(&Dcb->MainResource);
        }

        if (pSelf) {
            ExFreePoolWithTag(pSelf, EXT2_DENTRY_MAGIC);
        }
    }
    
    return Status;
}


NTSTATUS
Ext2TruncateBlock(
    IN PEXT2_IRP_CONTEXT IrpContext,
    IN PEXT2_VCB         Vcb,
    IN PEXT2_MCB         Mcb,
    IN ULONG             Layer,
    IN PULONG            BlockNo,
    IN PULONG            Extra
    )
{
    ULONG       i = 0;
    LONGLONG    Offset;

    PBCB        Bcb = NULL;
    PULONG      pData = NULL;

    NTSTATUS    Status = STATUS_SUCCESS;

    if (Layer == 0) {

        if (*BlockNo) {

            Status = Ext2FreeBlock(IrpContext, Vcb, *BlockNo);

            if (NT_SUCCESS(Status)) {
                ASSERT(Mcb->Inode.i_blocks >= BLOCK_SIZE / SECTOR_SIZE);
                if (Mcb->Inode.i_blocks < BLOCK_SIZE / SECTOR_SIZE) {
                    Mcb->Inode.i_blocks = 0;
                } else {
                    Mcb->Inode.i_blocks -= BLOCK_SIZE / SECTOR_SIZE;
                }
                *BlockNo = 0;
            }
        }

        if (Extra) {
            *Extra = *Extra - 1;
        }

    } else if (Layer <= 3) {

        if (*BlockNo) {

            Offset = (LONGLONG) (*BlockNo);
            Offset = Offset << BLOCK_BITS;

            if (!CcPinRead( Vcb->StreamObj,
                            (PLARGE_INTEGER) (&Offset),
                            BLOCK_SIZE,
                            PIN_WAIT,
                            &Bcb,
                            &pData )) {

                Ext2Print((DBG_ERROR, "Ext2TruncateBlock: CcPinReading error ...\n"));
                Status = STATUS_INSUFFICIENT_RESOURCES;
                goto errorout;
            }

            for (i = BLOCK_SIZE / 4; i > 0 && *Extra; i--) {
                Status = Ext2TruncateBlock(
                            IrpContext,
                            Vcb,
                            Mcb,
                            Layer - 1,
                            &pData[i-1],
                            Extra
                            );

                if (!NT_SUCCESS(Status)) {
                    break;
                }
            }

            if (i < BLOCK_SIZE / 4) {
                CcSetDirtyPinnedData(Bcb, NULL);
                Ext2RepinBcb(IrpContext, Bcb);
                Ext2AddMcbEntry(Vcb, Offset, (LONGLONG)BLOCK_SIZE);
            }

            if (i == 0) {
                Status = Ext2TruncateBlock(
                            IrpContext,
                            Vcb,
                            Mcb,
                            0,
                            BlockNo,
                            NULL
                            );
            }
        } else {
            *Extra -= Vcb->NumBlocks[Layer];
        }
    }

errorout:

    if (pData) {
        CcUnpinData(Bcb);
    }

    return Status;
}

BOOLEAN
Ext2AddMcbEntry (
    IN PEXT2_VCB Vcb,
    IN LONGLONG  Lba,
    IN LONGLONG  Length)
{
    BOOLEAN     bRet = FALSE;

    LONGLONG    Offset;

#if EXT2_DEBUG
    LONGLONG    DirtyLba;
    LONGLONG    DirtyLen;
#endif


    Offset = Lba & (~((LONGLONG)BLOCK_SIZE - 1));

    Length = (Length + Lba - Offset + BLOCK_SIZE - 1) &
             (~((LONGLONG)BLOCK_SIZE - 1));

    ASSERT ((Offset & (BLOCK_SIZE - 1)) == 0);
    ASSERT ((Length & (BLOCK_SIZE - 1)) == 0);

    Offset = (Offset >> BLOCK_BITS) + 1;
    Length = (Length >> BLOCK_BITS);

    ExAcquireResourceExclusiveLite(
        &(Vcb->McbResource),
        TRUE );

    __try {

        bRet = FsRtlAddLargeMcbEntry(
                         &(Vcb->DirtyMcbs),
                         Offset, Offset,
                         Length );

    } __except (EXCEPTION_EXECUTE_HANDLER) {

        DbgBreak();
        bRet = FALSE;
    }

#if EXT2_DEBUG
    if (bRet) {

        BOOLEAN     bFound = FALSE;
        LONGLONG    RunStart;
        LONGLONG    RunLength;
        ULONG       Index;

        bFound = FsRtlLookupLargeMcbEntry(
                        &(Vcb->DirtyMcbs),
                        Offset,
                        &DirtyLba,
                        &DirtyLen,
                        &RunStart,
                        &RunLength,
                        &Index );

        if ((!bFound) || (DirtyLba == -1) ||
            (DirtyLba != Offset) || (DirtyLen < Length)) {
            LONGLONG            sDirtyVba = 0;
            LONGLONG            sDirtyLba = 0;
            LONGLONG            sDirtyLength = 0;

            DbgBreak();

            for ( Index = 0; 
                  FsRtlGetNextLargeMcbEntry( &(Vcb->DirtyMcbs),
                                             Index,
                                             &sDirtyVba,
                                             &sDirtyLba,
                                             &sDirtyLength); 
                  Index++ ) {

                Ext2Print((DBG_INFO, "Index = %xh\n", Index));
                Ext2Print((DBG_INFO, "DirtyVba = %I64xh\n", sDirtyVba));
                Ext2Print((DBG_INFO, "DirtyLba = %I64xh\n", sDirtyLba));
                Ext2Print((DBG_INFO, "DirtyLen = %I64xh\n\n", sDirtyLength));
            }
        }
    }
#endif

    ExReleaseResourceLite(&(Vcb->McbResource));

    return bRet;
}

VOID
Ext2RemoveMcbEntry (
    IN PEXT2_VCB Vcb,
    IN LONGLONG  Lba,
    IN LONGLONG  Length)
{
    LONGLONG Offset;

    Offset = Lba & (~((LONGLONG)BLOCK_SIZE - 1));

    Length = (Length + Lba - Offset + BLOCK_SIZE - 1) &
             (~((LONGLONG)BLOCK_SIZE - 1));

    ASSERT(Offset == Lba);

    ASSERT ((Offset & (BLOCK_SIZE - 1)) == 0);
    ASSERT ((Length & (BLOCK_SIZE - 1)) == 0);

    Offset = (Offset >> BLOCK_BITS) + 1;
    Length = (Length >> BLOCK_BITS);

    Ext2Print((DBG_INFO, "Ext2RemoveMcbEntry: Lba=%I64xh Length=%I64xh\n",
                          Offset, Length ));

    ExAcquireResourceExclusiveLite(
        &(Vcb->McbResource),
        TRUE );

    __try {
        FsRtlRemoveLargeMcbEntry(
            &(Vcb->DirtyMcbs),
            Offset, Length      );

    } __except (EXCEPTION_EXECUTE_HANDLER) {
        DbgBreak();
    }

#if EXT2_DEBUG
    {
        BOOLEAN  bFound = FALSE;
        LONGLONG DirtyLba, DirtyLen;

        bFound = FsRtlLookupLargeMcbEntry(
                        &(Vcb->DirtyMcbs),
                        Offset,
                        &DirtyLba,
                        &DirtyLen,
                        NULL,
                        NULL,
                        NULL );

        if (bFound &&( DirtyLba != -1)) {
            DbgBreak();
        }
    }
#endif

    ExReleaseResourceLite(&(Vcb->McbResource));
}

BOOLEAN
Ext2LookupMcbEntry (
    IN PEXT2_VCB    Vcb,
    IN LONGLONG     Lba,
    OUT PLONGLONG   pLba,
    OUT PLONGLONG   pLength,
    OUT PLONGLONG   RunStart,
    OUT PLONGLONG   RunLength,
    OUT PULONG      Index)
{
    BOOLEAN     bRet;
    LONGLONG    Offset;


    Offset = Lba & (~((LONGLONG)BLOCK_SIZE - 1));
    ASSERT ((Offset & (BLOCK_SIZE - 1)) == 0);

    ASSERT(Lba == Offset);

    Offset = (Offset >> BLOCK_BITS) + 1;

    ExAcquireResourceSharedLite(
        &(Vcb->McbResource),
        TRUE );

    bRet = FsRtlLookupLargeMcbEntry(
                &(Vcb->DirtyMcbs),
                Offset,
                pLba,
                pLength,
                RunStart,
                RunLength,
                Index );

    ExReleaseResourceLite(&(Vcb->McbResource));

    if (bRet) {

        if (pLba && ((*pLba) != -1)) {

            ASSERT((*pLba) > 0);

            (*pLba) = (((*pLba) - 1) << BLOCK_BITS);
            (*pLba) += ((Lba) & ((LONGLONG)BLOCK_SIZE - 1));
        }

        if (pLength) {

            (*pLength) <<= BLOCK_BITS;
            (*pLength)  -= ((Lba) & ((LONGLONG)BLOCK_SIZE - 1));
        }

        if (RunStart && (*RunStart != -1)) {
            (*RunStart) = (((*RunStart) - 1) << BLOCK_BITS);
        }

        if (RunLength) {
            (*RunLength) <<= BLOCK_BITS;
        }
    }

    return bRet;
}

